# -*- coding: UTF-8 -*-
"""This module provides the BrowseRecord class."""


class BrowseRecord(object):
    """Base class that all browsable records inherit from.
    No attributes should be defined in this class (except _id/id, __oerp__,
    __osv__, __data__ and Python magic methods) in order to not be conflicted with the
    fields defined in the OSV model class on OpenERP server.

    """
    __oerp__ = None
    __osv__ = None
    def __init__(self, o_id):
        self._id = o_id
        self.__data__ = {'raw_data': False, 'fields_updated': []}

    @property
    def id(self):
        """ID of the record."""
        return self._id

    def __str__(self):
        return "browse_record({osv_name}, {obj_id})".format(
                osv_name=self.__osv__['name'],
                obj_id=self._id)

    def __repr__(self):
        return str(self)

    def __getitem__(self, key):
        return getattr(self, key)

    def __eq__(self, other):
        """Compare two browse records. Return ``True`` if their ID and OSV name
        are equals.

        NOTE: the comparison is made this way because their OSV classes can be
        differents objects.

        """
        return isinstance(other, BrowseRecord) and \
            self.id == other.id and \
            self.__osv__['name'] == other.__osv__['name'] # and \
            #self.__osv__['columns'].keys() == other.__osv__['columns'].keys():

# vim:expandtab:smartindent:tabstop=4:softtabstop=4:shiftwidth=4:



# -*- coding: UTF-8 -*-

import xmlrpclib
import socket

from oerplib.rpc import socket_netrpc, error


class ServiceXMLRPC(object):
    def __init__(self, name, url):
        self._name = name
        self._url = url
        self._sock = xmlrpclib.ServerProxy(self._url, allow_none=True)

    def __getattr__(self, method):
        def rpc_method(*args):
            try:
                sock_method = getattr(self._sock, method, False)
                return sock_method(*args)
            #TODO NEED TEST
            except xmlrpclib.Fault as exc:
                raise error.ConnectorError(
                        repr(exc.faultCode) + ': ' + exc.faultString)
            #TODO NEED TEST
            except xmlrpclib.Error as exc:
                raise error.ConnectorError(' - '.join(exc.args))
            except socket.error as exc:
                raise error.ConnectorError(exc.strerror)
        return rpc_method


class ServiceNetRPC(object):
    def __init__(self, name, server, port):
        self._name = name
        self._server = server
        self._port = port

    def __getattr__(self, method):
        def rpc_method(*args):
            try:
                sock = socket_netrpc.NetRPC()
                sock.connect(self._server, self._port)
                sock.send((self._name, method, )+args)
                result = sock.receive()
                sock.disconnect()
                return result
            #TODO NEED TEST
            except socket_netrpc.NetRPCError as exc:
                raise error.ConnectorError(unicode(exc))
            except socket.error as exc:
                raise error.ConnectorError(exc.strerror)
        return rpc_method

# vim:expandtab:smartindent:tabstop=4:softtabstop=4:shiftwidth=4:



# -*- coding: UTF-8 -*-
"""This module contains the NetRPC class which implements
the NetRPC protocol.

"""

import socket
import pickle
import cStringIO


class NetRPCError(BaseException):
    """Exception raised by the NetRPC class when an error occured."""
    def __init__(self, faultCode, faultString):
        self.faultCode = faultCode
        self.faultString = faultString
        self.args = (faultCode, faultString)


class NetRPC(object):
    """Low level class for NetRPC protocol."""
    def __init__(self, sock=None):
        if sock is None:
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        else:
            self.sock = sock
        self.sock.settimeout(120)

    def connect(self, host, port=False):
        if not port:
            buf = host.split('//')[1]
            host, port = buf.split(':')
        self.sock.connect((host, int(port)))

    def disconnect(self):
        self.sock.shutdown(socket.SHUT_RDWR)
        self.sock.close()

    def send(self, msg, exception=False, traceback=None):
        msg = pickle.dumps([msg,traceback])
        size = len(msg)
        self.sock.send('%8d' % size)
        self.sock.send(exception and "1" or "0")
        totalsent = 0
        while totalsent < size:
            sent = self.sock.send(msg[totalsent:])
            if sent == 0:
                raise NetRPCError("RuntimeError", "Socket connection broken")
            totalsent = totalsent + sent

    def receive(self):
        buf = ''
        while len(buf) < 8:
            chunk = self.sock.recv(8 - len(buf))
            if chunk == '':
                raise NetRPCError("RuntimeError", "Socket connection broken")
            buf += chunk
        size = int(buf)
        buf = self.sock.recv(1)
        if buf != "0":
            exception = buf
        else:
            exception = False
        msg = ''
        while len(msg) < size:
            chunk = self.sock.recv(size-len(msg))
            if chunk == '':
                raise NetRPCError("RuntimeError", "Socket connection broken")
            msg = msg + chunk
        msgio = cStringIO.StringIO(msg)
        unpickler = pickle.Unpickler(msgio)
        unpickler.find_global = None
        res = unpickler.load()

        if isinstance(res[0], BaseException):
            if exception:
                raise NetRPCError(res[0], res[1])
            raise res[0]
        else:
            return res[0]

# vim:expandtab:smartindent:tabstop=4:softtabstop=4:shiftwidth=4:



# -*- coding: UTF-8 -*-

class ConnectorError(BaseException):
    """Exception raised if the informations supplied
    to initiate the connector are wrong.
    """
    pass

# vim:expandtab:smartindent:tabstop=4:softtabstop=4:shiftwidth=4:



# -*- coding: UTF-8 -*-
"""Provides the ``ConnectorXMLRPC`` and ``ConnectorNetRPC`` classes."""

#import xmlrpclib, socket
import abc
#import time

#from oerplib.rpc import error, socket_netrpc, service
from oerplib.rpc import error, service


class Connector(object):
    """Connector base class defining the interface used
    to interact with an OpenERP server.
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self, server, port):
        self.server = server
        try:
            int(port)
        except ValueError:
            txt = "The port '{0}' is invalid. An integer is required."
            txt = error.format(port)
            raise error.ConnectorError(txt)
        else:
            self.port = port


class ConnectorXMLRPC(Connector):
    """Connector class using XMLRPC protocol."""
    def __init__(self, server, port):
        super(ConnectorXMLRPC, self).__init__(server, port)
        self._url = 'http://{server}:{port}/xmlrpc'.format(server=self.server,
                                                           port=self.port)

    def __getattr__(self, service_name):
        url = self._url + '/' + service_name
        srv = service.ServiceXMLRPC(service_name, url)
        setattr(self, service_name, srv)
        return srv


class ConnectorNetRPC(Connector):
    """Connector class using NetRPC protocol."""
    def __init__(self, server, port):
        super(ConnectorNetRPC, self).__init__(server, port)

    def __getattr__(self, service_name):
        srv = service.ServiceNetRPC(service_name, self.server, self.port)
        setattr(self, service_name, srv)
        return srv

# vim:expandtab:smartindent:tabstop=4:softtabstop=4:shiftwidth=4:



# -*- coding: UTF-8 -*-
"""This module contains all exceptions raised by `OERPLib` when an error
occurred.

"""
#import traceback, sys


class Error(Exception):
    def __init__(self, message, oerp_traceback=False):
        super(Error, self).__init__()
        self.oerp_traceback = oerp_traceback
        self.message = message
        #self.tb = sys.exc_info()[2]
    def __str__(self):
        #traceback.print_tb(self.tb)
        return u"{message}".format(message=self.message)


class RPCError(Error):
    """Exception raised when an error related to a RPC query occurs."""
    pass


class InternalError(Error):
    """Exception raised when an error occurred during an internal operation."""
    pass

# vim:expandtab:smartindent:tabstop=4:softtabstop=4:shiftwidth=4:



# -*- coding: UTF-8 -*-
"""The `OERPLib` module defines the :class:`OERP` class.

The :class:`OERP` class manage the client-side operations which are related to
an `OpenERP` server. You can use this one to write Python programs that performs
a variety of automated jobs that communicate with an `OpenERP` server.

"""

__author__ = 'ABF Osiell - Sebastien Alix'
__email__ = 'sebastien.alix@osiell.com'
__licence__ = 'LGPL v3'
__version__ = '0.5.3'

from oerplib.oerp import OERP
from oerplib import error

# vim:expandtab:smartindent:tabstop=4:softtabstop=4:shiftwidth=4:



# -*- coding: UTF-8 -*-
"""This module contains classes representing the fields supported by OpenObject.
A field is a Python descriptor which defines getter/setter methods for
its related attribute.
"""
import datetime

from oerplib import error
from oerplib.service.osv import browse

def is_int(value):
    if isinstance(value, bool):
        return False
    try:
        int(value)
        return True
    except ValueError:
        return False


class BaseField(object):
    """Field which all other fields inherit.
    Manage common metadata.
    """
    def __init__(self, osv, name, data):
        self.osv = osv
        self.name = name
        self.type = 'type' in data and data['type'] or False
        self.string = 'string' in data and data['string'] or False
        self.size = 'size' in data and data['size'] or False
        self.required = 'required' in data and data['required'] or False
        self.readonly = 'readonly' in data and data['readonly'] or False
        self.help = 'help' in data and data['help'] or False
        self.states = 'states' in data and data['states'] or False

    def __get__(self, instance, owner):
        pass

    def __set__(self, instance, value):
        pass

    def __str__(self):
        """Return a human readable string representation of the field."""
        attrs = ['string', 'relation', 'required', 'readonly', 'size', 'domain']
        attrs_rep = []
        for attr in attrs:
            if hasattr(self, attr):
                value = getattr(self, attr)
                if value:
                    if isinstance(value, basestring):
                        attrs_rep.append("{0}='{1}'".format(attr, value))
                    else:
                        attrs_rep.append("{0}={1}".format(attr, value))
        attrs_rep = ", ".join(attrs_rep)
        return "{0}({1})".format(self.type.upper(), attrs_rep)

    def check_value(self, value):
        """Check the validity of a value for the field."""
        #if self.readonly:
        #    raise error.Error(
        #        u"'{field_name}' field is readonly".format(
        #            field_name=self.name))
        if value and self.size and len(value) > self.size:
            raise ValueError(
                u"Lenght of the '{field_name}' is limited to {size}".format(
                    field_name=self.name,
                    size=self.size))
        if not value and self.required:
            raise ValueError(
                u"'{field_name}' field require a value".format(
                    field_name=self.name))
        return value


class SelectionField(BaseField):
    """Represent the OpenObject 'fields.selection'"""
    def __init__(self, osv, name, data):
        super(SelectionField, self).__init__(osv, name, data)
        self.selection = 'selection' in data and data['selection'] or False

    def __get__(self, instance, owner):
        return getattr(instance, "_{0}".format(self.name))

    def __set__(self, instance, value):
        value = self.check_value(value)
        setattr(instance, "_{0}".format(self.name), value)
        instance.__data__['fields_updated'].append(self.name)

    def check_value(self, value):
        super(SelectionField, self).check_value(value)
        selection = [val[0] for val in self.selection]
        if value and value not in selection:
            raise ValueError(
                u"The value '{value}' supplied doesn't match with the possible "
                u"values '{selection}' for the '{field_name}' field".format(
                    value=value,
                    selection=selection,
                    field_name=self.name,
            ))
        return value


class Many2ManyField(BaseField):
    """Represent the OpenObject 'fields.many2many'"""
    def __init__(self, osv, name, data):
        super(Many2ManyField, self).__init__(osv, name, data)
        self.relation = 'relation' in data and data['relation'] or False
        self.context = 'context' in data and data['context'] or False
        self.domain = 'domain' in data and data['domain'] or False

    def __get__(self, instance, owner):
        return [instance.__oerp__.browse(self.relation, o_id)
                for o_id in instance.__data__['raw_data'][self.name]]

    def __set__(self, instance, value):
        raise error.InternalError(u"Not implemented yet")

    def check_value(self, value):
        #TODO Many2Many.check_value()
        super(Many2ManyField, self).check_value(value)


class Many2OneField(BaseField):
    """Represent the OpenObject 'fields.many2one'"""
    def __init__(self, osv, name, data):
        super(Many2OneField, self).__init__(osv, name, data)
        self.relation = 'relation' in data and data['relation'] or False
        self.context = 'context' in data and data['context'] or False
        self.domain = 'domain' in data and data['domain'] or False

    def __get__(self, instance, owner):
        if getattr(instance, "_{0}".format(self.name)):
            return instance.__class__.__oerp__.browse(
                    self.relation,
                    getattr(instance, "_{0}".format(self.name))[0]
                )
        return False

    def __set__(self, instance, value):
        if isinstance(value, browse.BrowseRecord):
            o_rel = value
        elif is_int(value):
            o_rel = instance.__class__.__oerp__.browse(self.relation, value)
        elif value in [None, False]:
            o_rel = False
        else:
            raise ValueError(u"Value supplied has to be an integer, "
                             u"a browse_record object or False.")
        o_rel = self.check_value(o_rel)
        setattr(instance, "_{0}".format(self.name), o_rel and [o_rel.id, False])
        #setattr(instance, "_{0}".format(self.name), o_rel and o_rel.id)
        instance.__data__['fields_updated'].append(self.name)

    def check_value(self, value):
        super(Many2OneField, self).check_value(value)
        if value and value.__osv__['name'] != self.relation:
            raise ValueError(
                (u"Instance of '{osv_name}' supplied doesn't match with the " +\
                u"relation '{relation}' of the '{field_name}' field.").format(
                    osv_name=value.__osv__['name'],
                    relation=self.relation,
                    field_name=self.name))
        return value


class One2ManyField(BaseField):
    """Represent the OpenObject 'fields.one2many'"""
    def __init__(self, osv, name, data):
        super(One2ManyField, self).__init__(osv, name, data)
        self.relation = 'relation' in data and data['relation'] or False
        self.context = 'context' in data and data['context'] or False
        self.domain = 'domain' in data and data['domain'] or False

    def __get__(self, instance, owner):
        return [instance.__oerp__.browse(self.relation, o_id)
                for o_id in instance.__data__['raw_data'][self.name]]

    #def __set__(self, instance, value):

    #def check_value(self, value):
    #    super(One2ManyField, self).check_value(value)
    #    oerp = self.osv._oerp
    #    value_osv = oerp.pool.get_by_class(value.__class__)
    #    if value_osv.osv['name'] != self.relation:
    #        raise ValueError(
    #            (u"Instance of '{osv_name}' supplied doesn't match with the "+\
    #            u"relation '{relation}' of the '{field_name}' field.").format(
    #                osv_name=value_osv.osv['name'],
    #                relation=self.relation,
    #                field_name=self.name))
    #    return value


class DateField(BaseField):
    """Represent the OpenObject 'fields.data'"""
    pattern = "%Y-%m-%d"
    def __init__(self, osv, name, data):
        super(DateField, self).__init__(osv, name, data)

    def __get__(self, instance, owner):
        value = getattr(instance, "_{0}".format(self.name))
        try:
            res = datetime.datetime.strptime(value, self.pattern).date()
        except Exception: # ValueError, TypeError
            res = value
        return res

    def __set__(self, instance, value):
        value = self.check_value(value)
        setattr(instance, "_{0}".format(self.name), value)
        instance.__data__['fields_updated'].append(self.name)

    def check_value(self, value):
        super(DateField, self).check_value(value)
        if isinstance(value, datetime.date):
            value = value.strftime("%Y-%m-%d")
        elif isinstance(value, basestring):
            try:
                datetime.datetime.strptime(value, self.pattern)
            except:
                raise ValueError(
                    "String not well formatted, expecting '{0}' format".format(
                        self.pattern))
        elif isinstance(value, bool):
            return value
        else:
            raise ValueError(
                    "Expecting a datetime.date object or basestring")
        return value


class DateTimeField(BaseField):
    """Represent the OpenObject 'fields.datetime'"""
    pattern = "%Y-%m-%d %H:%M:%S"
    def __init__(self, osv, name, data):
        super(DateTimeField, self).__init__(osv, name, data)

    def __get__(self, instance, owner):
        value = getattr(instance, "_{0}".format(self.name))
        try:
            res = datetime.datetime.strptime(value, self.pattern)
        except Exception: # ValueError, TypeError
            res = value
        return res

    def __set__(self, instance, value):
        value = self.check_value(value)
        setattr(instance, "_{0}".format(self.name), value)
        instance.__data__['fields_updated'].append(self.name)

    def check_value(self, value):
        super(DateTimeField, self).check_value(value)
        if isinstance(value, datetime.datetime):
            value = value.strftime("%Y-%m-%d %H:%M:%S")
        elif isinstance(value, basestring):
            try:
                datetime.datetime.strptime(value, self.pattern)
            except:
                raise ValueError(
                    "Value not well formatted, expecting '{0}' format".format(
                        self.pattern))
        elif isinstance(value, bool):
            return value
        else:
            raise ValueError(
                    "Expecting a datetime.datetime object or basestring")
        return value


class ValueField(BaseField):
    """Represent simple OpenObject fields:
    - 'fields.char',
    - 'fields.float',
    - 'fields.integer',
    - 'fields.boolean',
    - 'fields.text',
    - 'fields.binary',
    """
    def __init__(self, osv, name, data):
        super(ValueField, self).__init__(osv, name, data)

    def __get__(self, instance, owner):
        return getattr(instance, "_{0}".format(self.name))

    def __set__(self, instance, value):
        value = self.check_value(value)
        setattr(instance, "_{0}".format(self.name), value)
        instance.__data__['fields_updated'].append(self.name)


def generate_field(osv, name, data):
    """Generate a well-typed field according to the data dictionary supplied
    (obtained via 'fields_get' XML-RPC/NET-RPC method).

    """
    assert 'type' in data
    field = None
    if data['type'] == 'selection':
        field = SelectionField(osv, name, data)
    elif data['type'] == 'many2many':
        field = Many2ManyField(osv, name, data)
    elif data['type'] == 'many2one':
        field = Many2OneField(osv, name, data)
    elif data['type'] == 'one2many':
        field = One2ManyField(osv, name, data)
    elif data['type'] == 'date':
        field = DateField(osv, name, data)
    elif data['type'] == 'datetime':
        field = DateTimeField(osv, name, data)
    elif data['type'] in ['char', 'float', 'integer',
                          'boolean', 'text', 'binary']:
        field = ValueField(osv, name, data)
    else:
        txt = (u"Can't instanciate the field '{field_name}', "
               u"'{field_type}' type unknown")
        raise error.InternalError(
            txt.format(field_name=name, field_type=data['type']))
    return field

# vim:expandtab:smartindent:tabstop=4:softtabstop=4:shiftwidth=4:



# -*- coding: UTF-8 -*-
"""This module provides a `RPC` connector (via the
:func:`get_connector <oerplib.rpc.get_connector>` method) which
use the `XML-RPC` or `Net-RPC` protocol to communicate with an `OpenERP` server.

Afterwards, `RPC` services (like `db`, `common` or `object`) and their methods
associated can be accessed dynamically from the connector returned.

Here is an example of how to use it:

    Get a `RPC` connector object::

    >>> from oerplib import rpc
    >>> cnt = rpc.get_connector('localhost', 8070, 'netrpc')

    Login and retrieve ID of the user connected::

    >>> uid = cnt.common.login('database', 'user', 'passwd')

    Execute a query::

    >>> res = cnt.object.execute('database', uid, 'passwd', 'res.partner', 'read', 42)

    Execute a workflow query::

    >>> res = cnt.object.exec_workflow('database', uid, 'passwd', 'sale.order', 'order_confirm', 4)

This sub-module is used by `OERPLib` to execute all queries while abstracting
the protocol used.

"""

from oerplib.rpc import connector, error

PROTOCOLS = {
        'xmlrpc': connector.ConnectorXMLRPC,
        'netrpc': connector.ConnectorNetRPC,
        }

def get_connector(server, port, protocol='xmlrpc'):
    """Return a `RPC` connector to interact with an `OpenERP` server.
    This one use either the `XML-RPC` protocol (by default) or `Net-RPC`,
    at the discretion of the user.
    """
    if protocol not in PROTOCOLS:
        txt = ("The protocol '{0}' is not supported. "
               "Please choose a protocol among these ones: {1}")
        txt = txt.format(protocol, PROTOCOLS.keys())
        raise error.ConnectorError(txt)
    return PROTOCOLS[protocol](server, port)

# vim:expandtab:smartindent:tabstop=4:softtabstop=4:shiftwidth=4:



# -*- coding: UTF-8 -*-
"""This module contains the ``OERP`` class which manage the interaction with
the `OpenERP` server.

"""
import os
import base64, zlib, tempfile
import time

from oerplib import rpc
from oerplib import error
from oerplib.service import db, osv


class OERP(object):
    """Return a new instance of the :class:`OERP` class.
    The optional `database` parameter specifies the default database to use
    when the :func:`login <oerplib.OERP.login>` method is called.
    If no `database` is set, the `database` parameter of the
    :func:`login <oerplib.OERP.login>` method will be mandatory.

    `XML-RPC` (by default, with the port 8069) and `Net-RPC` protocols
    are supported. Respective values for the `protocol` parameter are
    ``xmlrpc`` and ``netrpc``.

        >>> import oerplib
        >>> oerp = oerplib.OERP('localhost', protocol='netrpc', port=8070)

    :raise: :class:`oerplib.error.InternalError`

    """

    def __init__(self, server, database=None, protocol='xmlrpc', port=8069):
        self._server = server
        self._port = port
        self._protocol = protocol
        self._database = self._database_default = database
        self._pool = osv.Pool(self)
        self._user = None
        self._db = db.DB(self)
        # Instanciate the OpenERP server connector
        try:
            self._connector = rpc.get_connector(self._server, self._port,
                                                self._protocol)
        except rpc.error.ConnectorError as exc:
            raise error.InternalError(unicode(exc))

    # Readonly properties
    @property
    def user(self):
        """The browsable record of the user connected.

        >>> oerp.login('admin', 'admin') == oerp.user
        True

        """
        return self._user

    #user = property(lambda self: self._user,
    #                doc="Return the browsable record of the user connected.")
    server = property(lambda self: self._server,
                      doc="The server name.")
    port = property(lambda self: self._port,
                    doc="The port used.")
    protocol = property(lambda self: self._protocol,
                        doc="The protocol used.")
    database = property(lambda self: self._database,
                        doc="The database currently used.")
    db = property(lambda self: self._db,
                  doc=(""".. versionadded:: 0.4.0

                       The database management service.
                       See the :class:`oerplib.service.db.DB` class."""))

    #NOTE: in the past this function was implemented as a decorator for other
    # methods needed to be checked, but Sphinx documentation generator is not
    # able to auto-document decorated methods.
    def _check_logged_user(self):
        """Check if a user is logged. Otherwise, an error is raised."""
        if not self._user:
            raise error.Error(
                u"User login required.")

    def login(self, user, passwd, database=None):
        """Log in as the given `user` with the password `passwd` on the
        database `database` and return the corresponding user as a browsable
        record (from the ``res.users`` model).
        If `database` is not specified, the default one will be used instead.

        >>> user = oerp.login('admin', 'admin', database='db_name')
        >>> user.name
        u'Administrator'

        :return: the user connected as a browsable record
        :raise: :class:`oerplib.error.RPCError`

        """
        # Raise an error if no database was given
        self._database = database or self._database_default
        if not self._database:
            raise error.Error(u"No database specified")
        # Get the user's ID and generate the corresponding User record
        try:
            user_id = self._connector.common.login(self._database, user, passwd)
        except rpc.error.ConnectorError as exc:
            raise error.RPCError(unicode(exc))
        else:
            if user_id:
                #NOTE: create a fake User record just to execute the
                # first query : browse the real User record
                self._user = type('FakeUser', (osv.BrowseRecord,), {
                                    'id': None,
                                    'login': None,
                                    'password': None,
                                })
                self._user.id = user_id
                self._user.login = user
                self._user.password = passwd
                self._user = self.browse('res.users', user_id)
                return self._user
            else:
                #FIXME: Raise an error?
                raise error.RPCError(u"Wrong login ID or password")

    # ------------------------- #
    # -- Raw XML-RPC methods -- #
    # ------------------------- #

    def execute(self, osv_name, method, *args):
        """Execute a simple `XML-RPC` `method` on the OSV server class
        `osv_name`. `*args` parameters varies according to the `method` used.

        >>> oerp.execute('res.partner', 'read', [1, 2], ['name'])
        [{'name': u'ASUStek', 'id': 2}, {'name': u'Your Company', 'id': 1}]

        :return: the result returned by the `method` called
        :raise: :class:`oerplib.error.RPCError`

        """
        self._check_logged_user()
        # Execute the query
        try:
            return self._connector.object.execute(self._database, self._user.id,
                                                  self._user.password,
                                                  osv_name, method, *args)
        except rpc.error.ConnectorError as exc:
            raise error.RPCError(unicode(exc))

    def exec_workflow(self, osv_name, signal, obj_id):
        """`XML-RPC` Workflow query. Execute the workflow signal ``signal`` on
        the instance having the ID ``obj_id`` of OSV server class ``osv_name``.

        :raise: :class:`oerplib.error.RPCError`

        `WARNING: not sufficiently tested.`

        """
        #TODO NEED TEST
        self._check_logged_user()
        # Execute the workflow query
        try:
            self._connector.object.exec_workflow(self._database, self._user.id,
                                                 self._user.password,
                                                 osv_name, signal, obj_id)
        except rpc.error.ConnectorError as exc:
            raise error.RPCError(unicode(exc))

    def report(self, report_name, osv_name, obj_id, report_type='pdf',
               context=None):
        """Download a report from the `OpenERP` server and return
        the path of the file.

        >>> oerp.report('sale.order', 'sale.order', 1)
        '/tmp/oerplib_uJ8Iho.pdf'

        :return: the path to the generated temporary file
        :raise: :class:`oerplib.error.RPCError`

        """
        #TODO report_type: what it means exactly?

        self._check_logged_user()
        # If no context was supplied, get the default one
        context = context or self.execute('res.users', 'context_get')
        # Execute the report query
        try:
            pdf_data = self._get_report_data(report_name, osv_name, obj_id,
                                             report_type, context)
        except rpc.error.ConnectorError as exc:
            raise error.RPCError(unicode(exc))
        return self._print_file_data(pdf_data)

    def _get_report_data(self, report_name, osv_name, obj_id,
                         report_type='pdf', context=None):
        """Download binary data of a report from the `OpenERP` server."""
        if context is None:
            context = {}
        data = {'model': osv_name, 'id': obj_id, 'report_type': report_type}
        try:
            report_id = self._connector.report.report(
                            self._database, self.user.id, self.user.password,
                            report_name, [obj_id], data, context)
        except rpc.error.ConnectorError as exc:
            raise error.RPCError(unicode(exc))
        state = False
        attempt = 0
        while not state:
            try:
                pdf_data = self._connector.report.report_get(
                            self._database, self.user.id, self.user.password,
                            report_id)
            except rpc.error.ConnectorError as exc:
                raise error.RPCError("Unknown error occurred during the "
                                     "download of the report.")
            state = pdf_data['state']
            if not state:
                time.sleep(1)
                attempt += 1
            if attempt > 200:
                raise error.RPCError("Download time exceeded, "
                                     "the operation has been canceled.")
        return pdf_data

    @staticmethod
    def _print_file_data(data):
        """Print data in a temporary file and return the path of this one."""
        if 'result' not in data:
            raise error.InternalError(
                    u"Invalid data, the operation has been canceled.")
        content = base64.decodestring(data['result'])
        if data.get('code') == 'zlib':
            content = zlib.decompress(content)

        if data['format'] in ['pdf', 'html', 'doc', 'xls',
                              'sxw', 'odt', 'tiff']:
            if data['format'] == 'html' and os.name == 'nt':
                data['format'] = 'doc'
            (file_no, file_path) = tempfile.mkstemp('.'+data['format'],
                                                    'oerplib_')
            with file(file_path, 'wb+') as fp:
                fp.write(content)
            os.close(file_no)
            return file_path

    # ------------------------- #
    # -- High Level methods  -- #
    # ------------------------- #

    def browse(self, osv_name, ids, context=None):
        """Browse one record or several records (if ``ids`` is a list of IDs)
        according to the model ``osv_name``.

        >>> oerp.browse('res.partner', 1)
        browse_record(res.partner, 1)

        >>> [partner.name for partner in oerp.browse('res.partner', [1, 2])]
        [u'Your Company', u'ASUStek']

        :return: a BrowseRecord instance
        :return: a generator to iterate on BrowseRecord instances
        :raise: :class:`oerplib.error.RPCError`

        """
        return self.get(osv_name).browse(ids, context)

    def search(self, osv_name, args=None, offset=0, limit=None, order=None,
               context=None, count=False):
        """Return a list of IDs of records matching the given criteria in
        ``args`` parameter. ``args`` must be of the form
        ``[('name', '=', 'John'), (...)]``

        >>> oerp.search('res.partner', [('name', 'like', 'Agrolait')])
        [3]

        :return: a list of IDs
        :raise: :class:`oerplib.error.RPCError`

        """
        if args is None:
            args = []
        return self.execute(osv_name, 'search', args, offset, limit, order,
                            context, count)

    def create(self, osv_name, vals, context=None):
        """Create a new record with the specified values contained in the
        ``vals`` dictionary (e.g. ``{'name': 'John', ...}``).

        >>> partner_id = oerp.create('res.partner', {'name': 'Jacky Bob', 'lang': 'fr_FR'})

        :return: the ID of the new record.
        :raise: :class:`oerplib.error.RPCError`

        """
        return self.execute(osv_name, 'create', vals, context)

    def read(self, osv_name, ids, fields=None, context=None):
        """Return the ID of each record with the values
        of the requested fields ``fields`` from the OSV server class
        ``osv_name``. If ``fields`` is not specified, all fields values
        will be retrieved.

        >>> oerp.read('res.partner', [1, 2], ['name'])
        [{'name': u'ASUStek', 'id': 2}, {'name': u'Your Company', 'id': 1}]

        :raise: :class:`oerplib.error.RPCError`

        """
        if fields is None:
            fields = []
        return self.execute(osv_name, 'read', ids, fields, context)

    def write(self, osv_name, ids, vals=None, context=None):
        """Update records with given `ids` (e.g. ``[1, 42, ...]``)
        with the given values contained in the ``vals`` dictionary
        (e.g. ``{'name': 'John', ...}``).
        ``osv_name`` parameter is the OSV server class name
        (e.g. ``'sale.order'``).

        >>> oerp.write('res.users', [1], {'name': u"Administrator"})
        True

        :return: `True`
        :raise: :class:`oerplib.error.RPCError`

        """
        #if ids is None:
        #    ids = []
        if vals is None:
            vals = {}
        #if isinstance(osv_obj, osv.BrowseRecord):
        #    return self._pool.get_by_class(osv_obj.__class__).write(osv_obj)
        return self.execute(osv_name, 'write', ids, vals, context)

    def unlink(self, osv_name, ids, context=None):
        """Delete records with the given ``ids`` (e.g. ``[1, 42, ...]``).
        ``osv_name`` parameter is the OSV server class name
        (e.g. ``'sale.order'``).

        >>> oerp.unlink('res.partner', [1])

        :return: `True`
        :raise: :class:`oerplib.error.RPCError`

        """
        #if ids is None:
        #    ids = []
        #if isinstance(osv_obj, osv.BrowseRecord):
        #    return self._pool.get(osv_obj.__osv__['name']).unlink(osv_obj)
        return self.execute(osv_name, 'unlink', ids, context)

    # ---------------------- #
    # -- Special methods  -- #
    # ---------------------- #

    def write_record(self, browse_record, context=None):
        """.. versionadded:: 0.4.0

        Update the field values of ``browse_record`` by sending them to the
        `OpenERP` server (only field values which have been changed).

        """
        if not isinstance(browse_record, osv.BrowseRecord):
            raise ValueError(u"An instance of BrowseRecord is required")
        return self._pool.get_by_class(browse_record.__class__)._write_record(
                browse_record, context)

    def unlink_record(self, browse_record, context=None):
        """.. versionadded:: 0.4.0

        Delete the ``browse_record`` from the `OpenERP` server.

        """
        if not isinstance(browse_record, osv.BrowseRecord):
            raise ValueError(u"An instance of BrowseRecord is required")
        return self._pool.get_by_class(browse_record.__class__)._unlink_record(
                browse_record, context)

    def refresh(self, browse_record, context=None):
        """Restore original values of the ``browse_record`` from data
        retrieved on the OpenERP server.
        Thus, all changes made locally on the record are canceled.

        :raise: :class:`oerplib.error.RPCError`

        """
        return self._pool.get_by_class(browse_record.__class__)._refresh(
                browse_record, context)

    def reset(self, browse_record):
        """Cancel all changes made locally on the ``browse_record``.
        No request to the server is executed to perform this operation.
        Therefore, values restored may be outdated.

        """
        return self._pool.get_by_class(browse_record.__class__)._reset(
                browse_record)

    def get_osv_name(self, browse_record):
        """Return the OSV server class name of the ``browse_record`` supplied.

        >>> partner = oerp.browse('res.partner', 1)
        >>> oerp.get_osv_name(partner)
        'res.partner'

        :return: the OSV server class name of the browsable record

        """
        if not isinstance(browse_record, osv.BrowseRecord):
            raise ValueError(u"Value is not a browse_record.")
        return browse_record.__osv__['name']

    def get(self, osv_name):
        """.. versionadded:: 0.5.0

        Return a proxy of the OSV class `osv_name` built from the `OpenERP`
        server. See <TODO>.

        """
        return self._pool.get(osv_name)

# vim:expandtab:smartindent:tabstop=4:softtabstop=4:shiftwidth=4:



#!/usr/bin/env python
"""A sample script to demonstrate some of functionalities of OERPLib."""
import oerplib

# XMLRPC server configuration (NETRPC is also supported)
SERVER = 'localhost'
PROTOCOL = 'xmlrpc'
PORT = 8069
# Name of the OpenERP database to use
DATABASE = 'db_name'

USER = 'admin'
PASSWORD = 'password'

try:
    # Login
    oerp = oerplib.OERP(
            server=SERVER,
            database=DATABASE,
            protocol=PROTOCOL,
            port=PORT,
            )
    oerp.login(USER, PASSWORD)

    # ----------------------- #
    # -- Low level methods -- #
    # ----------------------- #

    # Execute - search
    user_ids = oerp.execute('res.users', 'search', [('id', '=', oerp.user.id)])

    # Execute - read
    user_data = oerp.execute('res.users', 'read', user_ids[0])

    # Execute - write
    oerp.execute('res.users', 'write', user_ids[0], {'name': u"Administrator"})

    # Execute - create
    new_user_id = oerp.execute('res.users', 'create', {'login': u"New user"})

    # ------------------------- #
    # -- Convenients methods -- #
    # ------------------------- #

    # Search IDs of a model that match criteria
    assert oerp.user.id in oerp.search('res.users', [('name', 'ilike', u"Administrator"),])

    # Create a record
    new_user_id = oerp.create('res.users', {'login': u"new_user"})

    # Read data of a record
    user_data = oerp.read('res.users', new_user_id)

    # Write a record
    oerp.write('res.users', [new_user_id], {'name': u"New user"})

    # Delete a record
    oerp.unlink('res.users', new_user_id)

    # -------------------- #
    # -- Browse objects -- #
    # -------------------- #

    # Browse an object
    user = oerp.browse('res.users', oerp.user.id)
    print(user.name)
    print(user.company_id.name)

    # .. or many objects
    for order in oerp.browse('sale.order', [68, 69]):
        print(order.name)
        print(order.partner_id.name)
        for line in order.order_line:
            print('\t{0}'.format(line.name))

    # ----------------------- #
    # -- Download a report -- #
    # ----------------------- #

    so_pdf_path = oerp.report('sale.order', 'sale.order', 1)
    inv_pdf_path = oerp.report('webkitaccount.invoice', 'account.invoice', 1)

    # -------------------- #
    # -- List databases -- #
    # -------------------- #

    # List databases
    print(oerp.db.list())

    # Create a database in background
    oerp.db.create(
            super_admin_passwd='super_admin_passwd',
            database='my_new_db',
            demo_data=True, lang='fr_FR',
            admin_passwd='admin_passwd')

    # Create a database (process blocked until the end of the operation)
    oerp.db.create_and_wait(
            super_admin_passwd='super_admin_passwd',
            database='my_new_db',
            demo_data=True, lang='fr_FR',
            admin_passwd='admin_passwd')

except oerplib.error.Error as e:
    print(e)
except Exception as e:
    print(e)




# -*- coding: UTF-8 -*-
"""This module contains the Pool class which manage OSV classes."""

from oerplib.service.osv import osv


class Pool(object):
    """Manage OSV classes."""

    def __init__(self, oerp):
        super(Pool, self).__init__()
        self.oerp = oerp

    def get(self, osv_name):
        """Return an OSV class which is able to create browsable records
        corresponding to the OSV name supplied.

        """
        return osv.OSV(self.oerp, osv_name)

    def get_by_class(self, browse_class):
        """Return a OSV class which is able to create browsable records
        corresponding to the BrowseRecord class supplied.

        """
        return osv.OSV(self.oerp, browse_class.__osv__['name'])

# vim:expandtab:smartindent:tabstop=4:softtabstop=4:shiftwidth=4:



# -*- coding: UTF-8 -*-
"""Provide the :class:`DB` class in order to manage the `OpenERP` databases."""

import time

from oerplib import rpc, error


class DB(object):
    """.. versionadded:: 0.4.0

    The `DB` class represents the database management service.
    It provides functionalities such as list, create, drop, dump
    and restore databases.

    .. note::
        This service have to be used through the :attr:`oerplib.OERP.db`
        property.

    >>> import oerplib
    >>> oerp = oerplib.OERP('localhost')
    >>> oerp.db
    <oerplib.service.db.DB object at 0xb75fb04c>

    .. warning::

        All methods documented below are not strictly implemented in `OERPLib`
        (except the
        :func:`create_and_wait <oerplib.service.db.DB.create_and_wait>` method).

        Method calls are purely dynamic, and the following documentation can be
        wrong if the API of `OpenERP` is changed between versions. Anyway, if
        you known the API used by the `OpenERP` server for the ``/db`` RPC
        service, it will work.

    .. method:: DB.list()

        Return a list of the `OpenERP` databases:

        >>> oerp.db.list()
        >>> ['production_db', 'test_db']

        :return: a list of database names

    .. method:: DB.list_lang()

        Return a list of codes and names of language supported by `OpenERP`:

        >>> oerp.db.list_lang()
        >>> [['sq_AL', u'Albanian / Shqipëri'], ['ar_AR', 'Arabic / الْعَرَبيّة'], ...]

        :return: a list of pairs representing languages with their codes and
                 names

    .. method:: DB.server_version()

        Return the version of the `OpenERP Server`:

        >>> oerp.db.server_version()
        >>> '6.1'

        :return: the version of the `OpenERP Server` as string

    .. method:: DB.dump(super_admin_passwd, database)

        Return a dump of `database` in `base64`:

        >>> binary_data = oerp.db.dump('super_admin_passwd', 'production_db')

        The super administrator password `super_admin_passwd` of `OpenERP` is
        required to perform this action.

        :return: the `base64` string representation of the `database`

    .. method:: DB.restore(super_admin_passwd, database, binary_data)

        Restore in `database` a dump previously created with the
        :func:`dump <DB.dump>` method:

        >>> oerp.db.restore('super_admin_passwd', 'test_db', binary_data)

        The super administrator password `super_admin_passwd` of `OpenERP` is
        required to perform this action.

    .. method:: DB.drop(super_admin_passwd, database)

        Drop the `database` from `OpenERP`:

        >>> oerp.db.drop('super_admin_passwd', 'test_db')
        True

        The super administrator password `super_admin_passwd` of `OpenERP` is
        required to perform this action.

        :return: True

    .. method:: DB.create(super_admin_passwd, database, demo_data=False, lang='en_US', admin_passwd='admin')

        Request the `OpenERP` server to create a new database named `database`
        which will have `admin_passwd` as administrator password and localized
        with the `lang` parameter.
        You have to set the flag `demo_data` to `True` in order to insert
        demonstration data.

        As the creating process may take some time, you can execute the
        :func:`get_process <DB.get_process>` method with the database ID
        returned to know its current state.

        >>> database_id = oerp.db.create('super_admin_passwd', 'test_db', False, 'fr_FR', 'my_admin_passwd')

        The super administrator password `super_admin_passwd` of `OpenERP` is
        required to perform this action.

        :return: the ID of the new database

    .. method:: DB.get_process(super_admin_passwd, database_id)

        Check the state of the creating process for the database identified by
        the `database_id` parameter.

        >>> oerp.db.get_process('super_admin_passwd', database_id) # Just after the call to the 'create' method
        (0, [])
        >>> oerp.db.get_process('super_admin_passwd', database_id) # Once the database is fully created
        (1.0, [{'login': u'admin', 'password': u'admin', 'name': u'Administrator'},
               {'login': u'demo', 'password': u'demo', 'name': u'Demo User'}])

        :return: A tuple with the progressing state and a list
                of user accounts created (once the database is fully created).

    """
    def __init__(self, oerp):
        self._oerp = oerp

    def create_and_wait(self, super_admin_passwd, database, demo_data=False,
                        lang='en_US', admin_passwd='admin'):
        """
        .. note::

            This method is not part of the official API of `OpenERP`. It's just
            a wrapper around the :func:`create <DB.create>` and
            :func:`get_process <DB.get_process>` methods.

        Like the :func:`create <DB.create>` method, but waits the end of
        the creating process by executing the
        :func:`get_process <DB.get_process>` method regularly to check its
        state.

        >>> oerp.db.create_and_wait('super_admin_passwd', 'test_db', False, 'fr_FR', 'my_admin_passwd')
        [{'login': u'admin', 'password': u'admin', 'name': u'Administrator'},
         {'login': u'demo', 'password': u'demo', 'name': u'Demo User'}]

        The super administrator password `super_admin_passwd` of `OpenERP` is
        required to perform this action.

        :return: a list of user accounts created
        :raise: :class:`oerplib.error.RPCError`

        """
        try:
            db_id = self._oerp._connector.db.create(
                    super_admin_passwd, database, demo_data, lang, admin_passwd)
            progress = 0.0
            attempt = 0
            while progress < 1.0:
                result = self._oerp._connector.db.get_progress(
                                super_admin_passwd, db_id)
                progress = result[0]
                if progress < 1.0:
                    time.sleep(1)
                    attempt += 1
                if attempt > 300:
                    raise error.RPCError(
                            "Too many attempts, the operation"
                            " has been canceled.")
            return result[1]

        except rpc.error.ConnectorError as exc:
            #FIXME handle the exception with the UnicodeEncodeError for
            # the error 'the database already exists'.
            #print dir(exc)
            raise error.RPCError(exc)

    def __getattr__(self, method):
        def rpc_method(*args):
            try:
                meth = getattr(self._oerp._connector.db, method, False)
                return meth(*args)
            except rpc.error.ConnectorError as exc:
                raise error.RPCError(exc)
        return rpc_method

# vim:expandtab:smartindent:tabstop=4:softtabstop=4:shiftwidth=4:



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
import os
from distutils.core import setup

name = 'OERPLib'
version = '0.5.3'
description = 'OpenERP client library which allows to easily interact with an OpenERP server.'
keywords = "openerp client xml-rpc xml_rpc xmlrpc net-rpc net_rpc netrpc oerplib communication lib library python service web webservice"
author = u"ABF Osiell - Sebastien Alix"
author_email = 'sebastien.alix@osiell.com'
url = 'http://packages.python.org/OERPLib/'
download_url = 'http://pypi.python.org/packages/source/O/OERPLib/OERPLib-%s.tar.gz' % version
license = 'LGPL v3'
doc_build_dir = 'doc/build'
doc_source_dir = 'doc/source'

cmdclass = {}
command_options = {}
# 'build_doc' option
try:
    from sphinx.setup_command import BuildDoc
    if not os.path.exists(doc_build_dir):
        os.mkdir(doc_build_dir)
    cmdclass = {'build_doc': BuildDoc}
    command_options = {
        'build_doc': {
            #'project': ('setup.py', name),
            'version': ('setup.py', version),
            'release': ('setup.py', version),
            'source_dir': ('setup.py', doc_source_dir),
            'build_dir': ('setup.py', doc_build_dir),
            'builder': ('setup.py', 'html'),
         }}
except Exception:
    print("No Sphinx module found. You have to install Sphinx "
          "to be able to generate the documentation.")

setup(name=name,
      version=version,
      description=description,
      long_description=open('README.txt').read(),
      keywords=keywords,
      author=author,
      author_email=author_email,
      url=url,
      download_url=download_url,
      packages=['oerplib',
                'oerplib.rpc',
                'oerplib.service',
                'oerplib.service.osv'],
      license=license,
      cmdclass=cmdclass,
      command_options=command_options,
      classifiers=[
          "Intended Audience :: Developers",
          "Programming Language :: Python",
          "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",
          "Topic :: Software Development :: Libraries :: Python Modules",
          ],
      )

# vim:expandtab:smartindent:tabstop=4:softtabstop=4:shiftwidth=4:



# -*- coding: UTF-8 -*-
"""Provide the :class:`OSV` class which allow to access dynamically to all methods
proposed by an OSV model of the `OpenERP` server."""

import collections

from oerplib import error
from oerplib.service.osv import fields, browse


class OSV(collections.Mapping):
    """.. versionadded:: 0.5.0

    Represent a data model from the `OpenERP` server.

    .. note::
        This class have to be used through the :func:`oerplib.OERP.get`
        method.

    >>> import oerplib
    >>> oerp = oerplib.OERP('localhost')
    >>> user = oerp.login('admin', 'passwd', 'database')
    >>> oerp.get('res.users')
    <oerplib.service.osv.osv.OSV object at 0x865f10c>

    .. warning::

        The only method implemented in this class is ``browse``. Except this
        one, method calls are purely dynamic. As long as you know the signature
        of the OSV method targeted, you will be able to use it.
        (see the tutorial `<REF NEEDED>`)

    """

    fields_reserved = ['id', '__oerp__', '__osv__', '__data__']
    def __init__(self, oerp, osv_name):
        super(OSV, self).__init__()
        self._oerp = oerp
        self._browse_class = self._generate_browse_class(osv_name)

    def _browse_generator(self, ids, context=None):
        """Generator used by the
        :func:`browse <oerplib.OERP.service.osv.osv.OSV.browse>` method.

        """
        for o_id in ids:
            yield self.browse(o_id, context)

    def browse(self, ids, context=None):
        """Browse one record or several records (if ``ids`` is a list of IDs)
        according to the model ``osv_name``.

        >>> oerp.get('res.partner').browse(1)
        browse_record(res.partner, 1)

        >>> [partner.name for partner in oerp.get('res.partner').browse([1, 2])]
        [u'Your Company', u'ASUStek']

        :return: a BrowseRecord instance
        :return: a generator to iterate on BrowseRecord instances
        :raise: :class:`oerplib.error.RPCError`

        """
        if isinstance(ids, list):
            return self._browse_generator(ids, context)
        else:
            obj = self._browse_class(ids)
            self._refresh(obj, context)
            return obj
            #return self.browse(ids, context)

    def _generate_browse_class(self, osv_name):
        """Generate a class with all its fields corresponding to
        the OSV name supplied and return them.

        """
        # Retrieve server fields info and generate corresponding local fields
        #FIXME try catch needed?
        try:
            fields_get = self._oerp.execute(osv_name, 'fields_get')
        except error.RPCError:
            raise error.RPCError(
                u"There is no OSV class named '{0}'.".format(osv_name))
        cls_name = osv_name.replace('.', '_')
        cls_fields = {}
        for field_name, field_data in fields_get.items():
            if field_name not in OSV.fields_reserved:
                cls_fields[field_name] = fields.generate_field(self,
                                                               field_name,
                                                               field_data)
        # Case where no field 'name' exists, we generate one (which will be
        # in readonly mode) in purpose to be filled with the 'name_get' method
        if 'name' not in cls_fields:
            field_data = {'type': 'text', 'string': 'Name', 'readonly': True}
            cls_fields['name'] = fields.generate_field(self, 'name', field_data)

        cls = type(cls_name, (browse.BrowseRecord,), {})
        cls.__oerp__ = self._oerp
        cls.__osv__ = {'name': osv_name, 'columns': cls_fields}
        slots = ['__oerp__', '__osv__', '__dict__', '__data__']
        slots.extend(cls_fields.keys())
        cls.__slots__ = slots
        return cls

    def _write_record(self, obj, context=None):
        """Send values of fields updated to the OpenERP server."""
        obj_data = obj.__data__
        vals = {}
        for field_name in obj_data['fields_updated']:
            if field_name in obj_data['raw_data']:
                field = self._browse_class.__osv__['columns'][field_name]
                # Many2One fields
                if isinstance(field, fields.Many2OneField):
                    value = getattr(obj, "_{0}".format(field_name))
                    vals[field_name] = value and value[0]
                # One2Many fields
                #elif isinstance(field, fields.One2ManyField):
                #    print field.relation
                #    for rel_id in getattr(obj, "_{0}".format(field_name)):
                #        self._oerp.write(field.relation, rel_id,
                #                        {'??FIELD??': obj.id})
                #    #vals[field_name] = getattr(obj, "_{0}".format(field_name))
                # Many2Many fields
                elif isinstance(field, fields.Many2ManyField):
                    pass #TODO write Many2Many value
                # All other fields
                else:
                    vals[field_name] = getattr(obj, "_{0}".format(field_name))
        try:
            res = self._oerp.write(obj.__osv__['name'], [obj.id], vals, context)
        except error.Error as exc:
            raise exc
        else:
            # Update raw_data dictionary
            self._refresh(obj, context) #FIXME delete to avoid a RPC request?
            return res

    def _refresh(self, obj, context=None):
        """Retrieve field values from OpenERP server.
        May be used to restore the original values
        in the purpose to cancel all changes made.

        """
        obj_data = obj.__data__
        obj_data['raw_data'] = self._oerp.read(obj.__osv__['name'], obj.id)
        if obj_data['raw_data'] is False:
            raise error.RPCError(
                u"There is no '{osv_name}' record with ID {obj_id}.".format(
                    osv_name=obj.__class__.__osv__['name'], obj_id=obj.id))
        # Special field 'name' have to be filled with the value returned
        # by the 'name_get' method
        #try:
        #    name = self._oerp.execute(obj.__osv__['name'], 'name_get', [obj.id])
        #except error.Error:
        #    pass
        #else:
        #    if name:
        #        try:
        #            obj_data['raw_data']['name'] = ast.literal_eval(name[0][1])
        #        except Exception:
        #            obj_data['raw_data']['name'] = name[0][1]
        self._reset(obj)

    def _reset(self, obj):
        """Cancel all changes by restoring field values with original values
        obtained during the last refresh (object instanciation or
        last call to _refresh() method).

        """
        obj_data = obj.__data__
        obj_data['fields_updated'] = []
        # Load fields and their values
        for field in self._browse_class.__osv__['columns'].values():
            if field.name in obj_data['raw_data']:
                setattr(obj, "_{0}".format(field.name),
                        obj_data['raw_data'][field.name])
                setattr(obj.__class__, field.name,
                        field)

    def _unlink_record(self, obj, context=None):
        """Delete the object from the OpenERP server."""
        return self._oerp.unlink(obj.__osv__, [obj.id], context)

    def __getattr__(self, method):
        def rpc_method(*args):
            result = self._oerp.execute(self._browse_class.__osv__['name'],
                                        method, *args)
            return result
        return rpc_method

    # ---------------------------- #
    # -- MutableMapping methods -- #
    # ---------------------------- #

    def __getitem__(self, obj_id):
        return self.browse(obj_id)

    def __iter__(self):
        ids = self._oerp.search(self._browse_class.__osv__['name'])
        for obj_id in ids:
            yield self.browse(obj_id)

    def __len__(self):
        return self._oerp.search(self._browse_class.__osv__['name'], count=True)

# vim:expandtab:smartindent:tabstop=4:softtabstop=4:shiftwidth=4:
